#!/usr/bin/env python
"""

Copyright (c) 2018 Alex Forencich

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from myhdl import *
import os

import axil

module = 'axil_dp_ram'
testbench = 'test_%s' % module

srcs = []

srcs.append("../rtl/%s.v" % module)
srcs.append("%s.v" % testbench)

src = ' '.join(srcs)

build_cmd = "iverilog -o %s.vvp %s" % (testbench, src)

def bench():

    # Parameters
    DATA_WIDTH = 32
    ADDR_WIDTH = 16
    STRB_WIDTH = int(DATA_WIDTH/8)
    PIPELINE_OUTPUT = 0

    # Inputs
    clk = Signal(bool(0))
    rst = Signal(bool(0))
    current_test = Signal(intbv(0)[8:])

    a_clk = Signal(bool(0))
    a_rst = Signal(bool(0))
    b_clk = Signal(bool(0))
    b_rst = Signal(bool(0))
    s_axil_a_awaddr = Signal(intbv(0)[ADDR_WIDTH:])
    s_axil_a_awprot = Signal(intbv(0)[3:])
    s_axil_a_awvalid = Signal(bool(0))
    s_axil_a_wdata = Signal(intbv(0)[DATA_WIDTH:])
    s_axil_a_wstrb = Signal(intbv(0)[STRB_WIDTH:])
    s_axil_a_wvalid = Signal(bool(0))
    s_axil_a_bready = Signal(bool(0))
    s_axil_a_araddr = Signal(intbv(0)[ADDR_WIDTH:])
    s_axil_a_arprot = Signal(intbv(0)[3:])
    s_axil_a_arvalid = Signal(bool(0))
    s_axil_a_rready = Signal(bool(0))
    s_axil_b_awaddr = Signal(intbv(0)[ADDR_WIDTH:])
    s_axil_b_awprot = Signal(intbv(0)[3:])
    s_axil_b_awvalid = Signal(bool(0))
    s_axil_b_wdata = Signal(intbv(0)[DATA_WIDTH:])
    s_axil_b_wstrb = Signal(intbv(0)[STRB_WIDTH:])
    s_axil_b_wvalid = Signal(bool(0))
    s_axil_b_bready = Signal(bool(0))
    s_axil_b_araddr = Signal(intbv(0)[ADDR_WIDTH:])
    s_axil_b_arprot = Signal(intbv(0)[3:])
    s_axil_b_arvalid = Signal(bool(0))
    s_axil_b_rready = Signal(bool(0))

    # Outputs
    s_axil_a_awready = Signal(bool(0))
    s_axil_a_wready = Signal(bool(0))
    s_axil_a_bresp = Signal(intbv(0)[2:])
    s_axil_a_bvalid = Signal(bool(0))
    s_axil_a_arready = Signal(bool(0))
    s_axil_a_rdata = Signal(intbv(0)[DATA_WIDTH:])
    s_axil_a_rresp = Signal(intbv(0)[2:])
    s_axil_a_rvalid = Signal(bool(0))
    s_axil_b_awready = Signal(bool(0))
    s_axil_b_wready = Signal(bool(0))
    s_axil_b_bresp = Signal(intbv(0)[2:])
    s_axil_b_bvalid = Signal(bool(0))
    s_axil_b_arready = Signal(bool(0))
    s_axil_b_rdata = Signal(intbv(0)[DATA_WIDTH:])
    s_axil_b_rresp = Signal(intbv(0)[2:])
    s_axil_b_rvalid = Signal(bool(0))

    # AXI4-Lite master
    axil_a_master_inst = axil.AXILiteMaster()
    axil_a_master_pause = Signal(bool(False))

    axil_a_master_logic = axil_a_master_inst.create_logic(
        a_clk,
        a_rst,
        m_axil_awaddr=s_axil_a_awaddr,
        m_axil_awprot=s_axil_a_awprot,
        m_axil_awvalid=s_axil_a_awvalid,
        m_axil_awready=s_axil_a_awready,
        m_axil_wdata=s_axil_a_wdata,
        m_axil_wstrb=s_axil_a_wstrb,
        m_axil_wvalid=s_axil_a_wvalid,
        m_axil_wready=s_axil_a_wready,
        m_axil_bresp=s_axil_a_bresp,
        m_axil_bvalid=s_axil_a_bvalid,
        m_axil_bready=s_axil_a_bready,
        m_axil_araddr=s_axil_a_araddr,
        m_axil_arprot=s_axil_a_arprot,
        m_axil_arvalid=s_axil_a_arvalid,
        m_axil_arready=s_axil_a_arready,
        m_axil_rdata=s_axil_a_rdata,
        m_axil_rresp=s_axil_a_rresp,
        m_axil_rvalid=s_axil_a_rvalid,
        m_axil_rready=s_axil_a_rready,
        pause=axil_a_master_pause,
        name='master_a'
    )

    axil_b_master_inst = axil.AXILiteMaster()
    axil_b_master_pause = Signal(bool(False))

    axil_b_master_logic = axil_b_master_inst.create_logic(
        b_clk,
        b_rst,
        m_axil_awaddr=s_axil_b_awaddr,
        m_axil_awprot=s_axil_b_awprot,
        m_axil_awvalid=s_axil_b_awvalid,
        m_axil_awready=s_axil_b_awready,
        m_axil_wdata=s_axil_b_wdata,
        m_axil_wstrb=s_axil_b_wstrb,
        m_axil_wvalid=s_axil_b_wvalid,
        m_axil_wready=s_axil_b_wready,
        m_axil_bresp=s_axil_b_bresp,
        m_axil_bvalid=s_axil_b_bvalid,
        m_axil_bready=s_axil_b_bready,
        m_axil_araddr=s_axil_b_araddr,
        m_axil_arprot=s_axil_b_arprot,
        m_axil_arvalid=s_axil_b_arvalid,
        m_axil_arready=s_axil_b_arready,
        m_axil_rdata=s_axil_b_rdata,
        m_axil_rresp=s_axil_b_rresp,
        m_axil_rvalid=s_axil_b_rvalid,
        m_axil_rready=s_axil_b_rready,
        pause=axil_b_master_pause,
        name='master_b'
    )

    # DUT
    if os.system(build_cmd):
        raise Exception("Error running build command")

    dut = Cosimulation(
        "vvp -m myhdl %s.vvp -lxt2" % testbench,
        clk=clk,
        rst=rst,
        current_test=current_test,

        a_clk=a_clk,
        a_rst=a_rst,
        b_clk=b_clk,
        b_rst=b_rst,
        s_axil_a_awaddr=s_axil_a_awaddr,
        s_axil_a_awprot=s_axil_a_awprot,
        s_axil_a_awvalid=s_axil_a_awvalid,
        s_axil_a_awready=s_axil_a_awready,
        s_axil_a_wdata=s_axil_a_wdata,
        s_axil_a_wstrb=s_axil_a_wstrb,
        s_axil_a_wvalid=s_axil_a_wvalid,
        s_axil_a_wready=s_axil_a_wready,
        s_axil_a_bresp=s_axil_a_bresp,
        s_axil_a_bvalid=s_axil_a_bvalid,
        s_axil_a_bready=s_axil_a_bready,
        s_axil_a_araddr=s_axil_a_araddr,
        s_axil_a_arprot=s_axil_a_arprot,
        s_axil_a_arvalid=s_axil_a_arvalid,
        s_axil_a_arready=s_axil_a_arready,
        s_axil_a_rdata=s_axil_a_rdata,
        s_axil_a_rresp=s_axil_a_rresp,
        s_axil_a_rvalid=s_axil_a_rvalid,
        s_axil_a_rready=s_axil_a_rready,
        s_axil_b_awaddr=s_axil_b_awaddr,
        s_axil_b_awprot=s_axil_b_awprot,
        s_axil_b_awvalid=s_axil_b_awvalid,
        s_axil_b_awready=s_axil_b_awready,
        s_axil_b_wdata=s_axil_b_wdata,
        s_axil_b_wstrb=s_axil_b_wstrb,
        s_axil_b_wvalid=s_axil_b_wvalid,
        s_axil_b_wready=s_axil_b_wready,
        s_axil_b_bresp=s_axil_b_bresp,
        s_axil_b_bvalid=s_axil_b_bvalid,
        s_axil_b_bready=s_axil_b_bready,
        s_axil_b_araddr=s_axil_b_araddr,
        s_axil_b_arprot=s_axil_b_arprot,
        s_axil_b_arvalid=s_axil_b_arvalid,
        s_axil_b_arready=s_axil_b_arready,
        s_axil_b_rdata=s_axil_b_rdata,
        s_axil_b_rresp=s_axil_b_rresp,
        s_axil_b_rvalid=s_axil_b_rvalid,
        s_axil_b_rready=s_axil_b_rready
    )

    @always(delay(4))
    def clkgen():
        clk.next = not clk
        a_clk.next = not a_clk
        b_clk.next = not b_clk

    def wait_normal():
        while not axil_a_master_inst.idle() or not axil_b_master_inst.idle():
            yield clk.posedge

    def wait_pause_master():
        while not axil_a_master_inst.idle() or not axil_b_master_inst.idle():
            axil_a_master_pause.next = True
            axil_b_master_pause.next = True
            yield clk.posedge
            yield clk.posedge
            yield clk.posedge
            axil_a_master_pause.next = False
            axil_b_master_pause.next = False
            yield clk.posedge

    @instance
    def check():
        yield delay(100)
        yield clk.posedge
        rst.next = 1
        a_rst.next = 1
        b_rst.next = 1
        yield clk.posedge
        rst.next = 0
        a_rst.next = 0
        b_rst.next = 0
        yield clk.posedge
        yield delay(100)
        yield clk.posedge

        # testbench stimulus

        yield clk.posedge
        print("test 1: read and write, port A")
        current_test.next = 1

        addr = 4
        test_data = b'\x11\x22\x33\x44'

        axil_a_master_inst.init_write(addr, test_data)

        yield axil_a_master_inst.wait()
        yield clk.posedge

        axil_a_master_inst.init_read(addr, len(test_data))

        yield axil_a_master_inst.wait()
        yield clk.posedge

        data = axil_a_master_inst.get_read_data()
        assert data[0] == addr
        assert data[1] == test_data

        yield delay(100)

        yield clk.posedge
        print("test 2: read and write, port B")
        current_test.next = 2

        addr = 4
        test_data = b'\x11\x22\x33\x44'

        axil_b_master_inst.init_write(addr, test_data)

        yield axil_b_master_inst.wait()
        yield clk.posedge

        axil_b_master_inst.init_read(addr, len(test_data))

        yield axil_b_master_inst.wait()
        yield clk.posedge

        data = axil_b_master_inst.get_read_data()
        assert data[0] == addr
        assert data[1] == test_data

        yield delay(100)

        yield clk.posedge
        print("test 3: various reads and writes on port A")
        current_test.next = 3

        for length in range(1,8):
            for offset in range(4,8):
                for wait in wait_normal, wait_pause_master:
                    print("length %d, offset %d"% (length, offset))
                    addr = 256*(16*offset+length)+offset
                    test_data = b'\x11\x22\x33\x44\x55\x66\x77\x88'[0:length]

                    axil_a_master_inst.init_write(addr-4, b'\xAA'*(length+8))

                    yield axil_a_master_inst.wait()

                    axil_a_master_inst.init_write(addr, test_data)

                    yield wait()

                    axil_a_master_inst.init_read(addr-1, length+2)

                    yield axil_a_master_inst.wait()

                    data = axil_a_master_inst.get_read_data()
                    assert data[0] == addr-1
                    assert data[1] == b'\xAA'+test_data+b'\xAA'

        for length in range(1,8):
            for offset in range(4,8):
                for wait in wait_normal, wait_pause_master:
                    print("length %d, offset %d"% (length, offset))
                    addr = 256*(16*offset+length)+offset
                    test_data = b'\x11\x22\x33\x44\x55\x66\x77\x88'[0:length]

                    axil_a_master_inst.init_write(addr, test_data)

                    yield axil_a_master_inst.wait()

                    axil_a_master_inst.init_read(addr, length)

                    yield wait()
                    yield clk.posedge

                    data = axil_a_master_inst.get_read_data()
                    assert data[0] == addr
                    assert data[1] == test_data

        yield delay(100)

        yield clk.posedge
        print("test 4: various reads and writes on port B")
        current_test.next = 4

        for length in range(1,8):
            for offset in range(4,8):
                for wait in wait_normal, wait_pause_master:
                    print("length %d, offset %d"% (length, offset))
                    addr = 256*(16*offset+length)+offset
                    test_data = b'\x11\x22\x33\x44\x55\x66\x77\x88'[0:length]

                    axil_b_master_inst.init_write(addr-4, b'\xAA'*(length+8))

                    yield axil_b_master_inst.wait()

                    axil_b_master_inst.init_write(addr, test_data)

                    yield wait()

                    axil_b_master_inst.init_read(addr-1, length+2)

                    yield axil_b_master_inst.wait()

                    data = axil_b_master_inst.get_read_data()
                    assert data[0] == addr-1
                    assert data[1] == b'\xAA'+test_data+b'\xAA'

        for length in range(1,8):
            for offset in range(4,8):
                for wait in wait_normal, wait_pause_master:
                    print("length %d, offset %d"% (length, offset))
                    addr = 256*(16*offset+length)+offset
                    test_data = b'\x11\x22\x33\x44\x55\x66\x77\x88'[0:length]

                    axil_b_master_inst.init_write(addr, test_data)

                    yield axil_b_master_inst.wait()

                    axil_b_master_inst.init_read(addr, length)

                    yield wait()
                    yield clk.posedge

                    data = axil_b_master_inst.get_read_data()
                    assert data[0] == addr
                    assert data[1] == test_data

        yield delay(100)

        yield clk.posedge
        print("test 5: arbitration test")
        current_test.next = 5

        for k in range(10):
            axil_a_master_inst.init_write(k*256, b'\x11\x22\x33\x44')
            axil_a_master_inst.init_read(k*256, 4)
            axil_b_master_inst.init_write(k*256, b'\x11\x22\x33\x44')
            axil_b_master_inst.init_read(k*256, 4)

        yield wait_normal()

        for k in range(10):
            axil_a_master_inst.get_read_data()
            axil_b_master_inst.get_read_data()

        yield delay(100)

        raise StopSimulation

    return instances()

def test_bench():
    sim = Simulation(bench())
    sim.run()

if __name__ == '__main__':
    print("Running test...")
    test_bench()
